'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const not = async (x) => {
     var operand = await x;
    return !operand
 };

const identity = x => x;

// predicates
const isEven = x => x % 2 === 0;
const isGreaterThan = x => y => y > x ? y : 0;
const digitize = x => x ? 1 : 0;

const modulus = val => x => x % val;

const select = property => x => x[property];

function apply (x, f) {
    return f(x)
}

function compose (...funcs) {
    return x => funcs.reduceRight(apply, x)
}

function concat (accumlator, val) {
    return accumlator.concat(val)
}

function latest (accumlator, val) {
    return val
}

async function transduceAsyncIterator (transform, reducerfunction, init, asynciterator) {
    var reducer = transform(reducerfunction);
    var n = null;
    do {
        n = await asynciterator.next();
        if (!n.done) {
        var v = n.value;
        init = await reducer(init, v);
        }
    } while (!n.done)
    return init
}

async function transduceAsyncHasNextIterator (transform, reducerfunction, init, asynchasnextiterator) {
    var reducer = transform(reducerfunction);
    var n = null;
    var r = null;
    do {
        n = await asynchasnextiterator.hasNext();
        if (n) {
            var v = await asynchasnextiterator.next();
            r = await reducer(init, v);
            if (r.hasOwnProperty('reduced')) {
                if (r.reduced) {
                    init = r.reduced;
                }
            } else if (r) {
                init = r;
            }
        }
    } while (r && n)
    return init
}

async function transduceArray (xf, rf, init, xs) {
    // call reduce on the data structure internally (abstract it away)
    // pass the rf to the composed transformation
    // pass in the initial value
    var xrf = await xf(rf);
    return xs.reduce(xrf, init)
}

async function * transduceGenerator (transform, reducerfunction, init, streamgenerator) {
    var reducer = transform(reducerfunction);
    for await (const value of streamgenerator) {
        var newinit = await reducer(init, value);
        // Here we checked if there is new a 'reduced' value and only generate a new value when this is the case
        if (!newinit) ; else if (newinit === init) ; else {
            if (newinit.hasOwnProperty('reduced')) {
                if (Array.isArray(newinit.reduced)) {
                    for (var i = 0; i < newinit.reduced.length; i++) {
                        init = newinit.reduced[i];
                        yield newinit.reduced[i];
                    }
                }
            } else {
                init = newinit;
                yield newinit;
            }
        }
    }
}

function passthrough () {
  return function (rf) {
    // this takes 2 things and makes them 1
    return async (acc, val) => {
      return rf(acc, val) // <-- rf replaces 'concat'
    }
  }
}

// generalize the 'mapping' concept, without the concat...
function mapping (f) {
    return function (rf) {
      // this takes 2 things and makes them 1
      return async (acc, val) => {
        var m = await f(val);
        var r = await rf(acc, m); // <-- rf replaces 'concat'
        if (r.hasOwnProperty('reduced')) {
          if (r.reduced) {
            return r
          } else {
            return acc
          }
        } else {
          return r
        }
      }
    }
  }

  // generalize the 'filtering' concept, without the concat...
  function filtering (p) {
    return function (rf) {
      // this takes 2 things and makes them 1
      return async (acc, val) => {
        var pred = (await p(val));
        return pred ? rf(acc, val) : acc // <-- rf replaces 'concat'
      }
    }
  }

function take (cnt) {
  // assert(cnt && cnt > 0)
  var count = cnt;
  return function (rf) {
    // this takes 2 things and makes them 1
    return async (acc, val) => {
      if ((--count) < 0) {
        return {
          reduced: null
        }
      } else {
        return rf(acc, val)
      }
    }
  }
}

function skip (cnt) {
  // assertMod(cnt && cnt >= 0)
  var count = 0;
  return function (rf) {
    // this takes 2 things and makes them 1
    return async (acc, val) => {
      if (count++ < cnt) {
        return acc
      } else {
        return rf(acc, val)
      }
    }
  }
}

function sampling (period) {
  // assertMod(cnt && cnt >= 0)
  var last = 0;
  return function (rf) {
    return async (acc, val) => {
      var nw = Date.now();
      var diff = nw - last;
      // console.log(diff)
      if (diff < period) {
        // console.log('Skip' + val)
        return acc
      } else {
        last = nw;
        // console.log('Accept' + val)
        return rf(acc, val)
      }
    }
  }
}

function eventing (p) {
  var latest = null;
  var sequence = -1;
  return function (rf) {
    return async (acc, val) => {
      sequence++;
      var pred = (await p(val));
      if (pred && !latest) {
        latest = {
          start: val.time ? val.time : sequence,
          end: val.time ? val.time : sequence
        };
        return acc
      } else if (pred && latest) {
        latest.end = val.time ? val.time : sequence;
        return acc
      } else if (!pred && latest) {
        var next = latest;
        latest = null;
        return rf(acc, next)
      } else {
        return acc
      }
    }
  }
}

/*
 *  Split one event into multiple
 *  f is a function which maps a value to an array
 */
function split (f) {
  return function (rf) {
    // this takes 2 things and makes them 1
    return async (acc, val) => {
      var rs = await f(val);
      var reduction = { reduced: [] };
      try {
        var acc2 = acc;
        for (var i = 0; i < rs.length; i++) {
          var r = rs[i];
          acc2 = await rf(acc2, r);
          if (acc2.hasOwnProperty('reduced')) {
            if (acc2.reduced) {
              reduction.reduced.push(acc2.reduced);
            }
          } else {
            reduction.reduced.push(acc2);
          }
        }
        return reduction
      } catch (ex) {
        console.log(ex);
      }
    }
  }
}

async function * makeAsyncRangeIterator (start = 0, end = Infinity, step = 1, delay = 10) {
    let iterationCount = 0;
    for (let i = start; i < end; i += step) {
        iterationCount++;
        yield new Promise((resolve, reject) => {
            setTimeout(() => {
                resolve(i);
            }, delay);
        });
    }
    return iterationCount
  }

function makeAsyncHasNextRangeIterator (start = 0, end = Infinity, step = 1) {
    let iterationCount = start;
    return {
        hasNext: function () {
            return iterationCount < end
        },
        next: function () {
            return new Promise((resolve, reject) => {
                setTimeout(() => {
                    resolve(iterationCount++);
                }, 10);
            })
        }
    }
}

function makeArrayIterator (dataArray) {
    let iterationCount = 0;
    return {
        hasNext: function () {
            return iterationCount < dataArray.length
        },
        next: function () {
            return dataArray[iterationCount++]
            }
        }
    }

exports.not = not;
exports.identity = identity;
exports.isEven = isEven;
exports.isGreaterThan = isGreaterThan;
exports.digitize = digitize;
exports.modulus = modulus;
exports.select = select;
exports.compose = compose;
exports.concat = concat;
exports.latest = latest;
exports.transduceAsyncIterator = transduceAsyncIterator;
exports.transduceAsyncHasNextIterator = transduceAsyncHasNextIterator;
exports.transduceArray = transduceArray;
exports.transduceGenerator = transduceGenerator;
exports.mapping = mapping;
exports.filtering = filtering;
exports.take = take;
exports.skip = skip;
exports.eventing = eventing;
exports.sampling = sampling;
exports.passthrough = passthrough;
exports.split = split;
exports.makeArrayIterator = makeArrayIterator;
exports.makeAsyncRangeIterator = makeAsyncRangeIterator;
exports.makeAsyncHasNextRangeIterator = makeAsyncHasNextRangeIterator;
//# sourceMappingURL=funprog.js.map
